#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = (
    "ProjectVisibility",
    "ProjectMemberRole",
    "MemberRole",
    "WorkspaceMemberRole",
)


class ProjectVisibility:
    PRIVATE = "priv"
    PUBLIC = "pub"
    WORKSPACE = "workspace"


class ProjectMemberRole:
    VIEWER = "viewer"
    OWNER = "owner"
    CONTRIBUTOR = "contributor"

    # Deprecated
    MEMBER = CONTRIBUTOR
    MANAGER = OWNER


MemberRole = ProjectMemberRole


class WorkspaceMemberRole:
    ADMIN = "admin"
    MEMBER = "member"
